//
// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
// Copyright (C) 2020 - Stephane MOTTELET
//
// This file is hereby licensed under the terms of the GNU GPL v2.0,
// For more information, see the COPYING file which you should have received
//
//--------------------------------------------------------------------------

#include "OptimizationManager.hxx"
#include "argumentsPositions.hxx"

// OptimizationManager methods

OptimizationManager::OptimizationManager(const std::wstring& callerName)
{
    m_wstrCaller = callerName;
    m_pstrCaller = wide_string_to_UTF8(callerName.c_str());
    m_iNbVariables = 0;
    m_iNbConstraints = 0;
    m_iNbConstraintsJacobianNonZeroTerms = 0;
    m_iNbHessianNonZeroTerms = 0;
    
    m_pdblSolution = NULL;
    m_pdblLambda = NULL;
    m_pdblZL = NULL;
    m_pdblZU = NULL;
    m_piNonLinearVariables = NULL;
    
    // Scilab internal types
    m_pDblX0 = NULL;
    m_pDblVariablesLinearity = NULL;
    m_pDblXLower = NULL;
    m_pDblXUpper = NULL;
    m_pDblConstraintsLhs = NULL;
    m_pDblConstraintsRhs = NULL;
    m_pDblConstraintsJacPattern = NULL;
    m_pDblHessianLowerTriPattern = NULL;
    m_pDblConstraintsLinearity = NULL;

    // arrays
    m_pCallFunction.fill(NULL);
    m_pEntryPointFunction.fill(NULL);
    m_pdblOutput.fill(NULL);
    m_iSizeOfOutput.fill(0);
    m_pCallFunctionName.fill(NULL);
    m_iNbFunctionCount.fill(0);
}

bool OptimizationManager::parseMatrices(types::typed_list &in)
{
    std::list<int> iMatrices = {X_POS, SPARSE_DGOBJ_POS,SPARSE_DHOBJ_POS,VAR_LIN_TYPE_POS,CONSTR_LIN_TYPE_POS,
        CONSTR_RHS_POS,CONSTR_LHS_POS,X_LOWER_POS,X_UPPER_POS};

    // basic test
    for (int i : iMatrices)
    {
        if (in[i]->isDouble() == false || in[i]->getAs<types::Double>()->isComplex())
        {
            Scierror(999, _("%s: Wrong type for input argument #%d (%s): A real matrix expected.\n"), getCallerStr(), i+1, pstrArgName[i]);
            return false;    
        }
    }

    if (in[X_POS]->getAs<types::Double>()->isEmpty())
    {
        Scierror(999, _("%s: Wrong size for input argument #%d (%s): Non-empty matrix expected.\n"), getCallerStr(), X_POS, pstrArgName[X_POS]);
        return false;    
    }
    else
    {
        m_pDblX0 = in[X_POS]->getAs<types::Double>();
        m_iNbVariables = m_pDblX0->getSize();
    }
    
    for (int i : {VAR_LIN_TYPE_POS, X_LOWER_POS, X_UPPER_POS})
    {
        if (in[i]->getAs<types::Double>()->isEmpty() == false && in[i]->getAs<types::Double>()->getSize() != m_iNbVariables)
        {
            Scierror(999,_("%s: Wrong size for input argument #%d (%s): A vector of size %d expected.\n"), getCallerStr(), i+1, pstrArgName[i], m_iNbVariables);
            return false;    
        }  
    }
    
    // Verify vectors and matrices
    
    m_pDblXLower                 = in[X_LOWER_POS]->getAs<types::Double>();
    m_pDblXUpper                 = in[X_UPPER_POS]->getAs<types::Double>();
    m_pDblVariablesLinearity     = in[VAR_LIN_TYPE_POS]->getAs<types::Double>();
    m_pDblConstraintsLhs         = in[CONSTR_LHS_POS]->getAs<types::Double>();
    m_pDblConstraintsRhs         = in[CONSTR_RHS_POS]->getAs<types::Double>();
    m_pDblConstraintsLinearity   = in[CONSTR_LIN_TYPE_POS]->getAs<types::Double>();
    m_pDblConstraintsJacPattern  = in[SPARSE_DGOBJ_POS]->getAs<types::Double>();
    m_pDblHessianLowerTriPattern = in[SPARSE_DHOBJ_POS]->getAs<types::Double>();
    
    int iCLhs = m_pDblConstraintsLhs->getSize();
    int iCRhs = m_pDblConstraintsRhs->getSize();
        
    m_iNbConstraints = std::max(iCLhs, iCRhs);
    if (m_iNbConstraints > 0)
    {
        if (std::min(iCLhs, iCRhs) > 0 && iCLhs != iCRhs)
        {
            Scierror(999,_("%s: Arguments #%d (%s) and #%d (%s) must have the same sizes.\n"), getCallerStr(), 
            pstrArgName[CONSTR_LHS_POS], CONSTR_LHS_POS+1, pstrArgName[CONSTR_RHS_POS], CONSTR_RHS_POS+1);
            return false;    
        }
        if (in[GOBJ_POS]->isDouble()) // i.e. is an empty matrix
        {
            Scierror(999,_("%s: parameter %d (%s) should be a function or a list.\n"), getCallerStr(), GOBJ_POS+1, pstrArgName[GOBJ_POS]);
            return false;    
        }
    }
    else if (in[GOBJ_POS]->isDouble() == false) // i.e. is not a empty matrix hence is a function or a list
    {
        Scierror(999,_("%s: Arguments #%d (%s) and #%d (%s) cannot be both empty.\n"), getCallerStr(), CONSTR_LHS_POS+1, 
        pstrArgName[CONSTR_LHS_POS], CONSTR_RHS_POS+1, pstrArgName[CONSTR_RHS_POS]);
        return false;    
    }

    // Verify constraints jacobian
    if (m_iNbConstraints > 0)
    {
        int iNbMaxTerms = (int) m_iNbVariables*m_iNbConstraints;

        if (m_pDblConstraintsJacPattern->isEmpty())
        {
            m_iNbConstraintsJacobianNonZeroTerms = iNbMaxTerms; // full pattern will be set in OptimizationManager::getConstraintsJacobianPattern
        }
        else if (m_pDblConstraintsJacPattern->getCols() != 2)
        {
            Scierror(999,_("%s: Wrong size for input argument #%d (%s): A matrix with %d columns expected.\n"), getCallerStr(), 
            SPARSE_DGOBJ_POS+1, pstrArgName[SPARSE_DGOBJ_POS], 2);            
            return false;    
        }
        else if (m_pDblConstraintsJacPattern->getRows() > iNbMaxTerms)
        {
            Scierror(999,_("%s: Wrong size for input argument #%d (%s): A matrix with at most %d rows expected.\n"), getCallerStr(), 
            SPARSE_DGOBJ_POS+1, pstrArgName[SPARSE_DGOBJ_POS], iNbMaxTerms);            
            return false;    
        }
        else
        {
            for (int i=0; i<m_pDblConstraintsJacPattern->getRows(); i++)
            {
                int iCstr = m_pDblConstraintsJacPattern->get(i,0);
                int iVar = m_pDblConstraintsJacPattern->get(i,1);
                if (iCstr < 1 || iCstr > m_iNbConstraints || iVar < 1 || iVar > m_iNbVariables)
                {
                    Scierror(999,_("%s: invalid value in argument #%d (%s) at row %d.\n"), getCallerStr(), SPARSE_DGOBJ_POS+1, pstrArgName[SPARSE_DGOBJ_POS], i+1);    
                    return false;                            
                }
            }
            m_iNbConstraintsJacobianNonZeroTerms = m_pDblConstraintsJacPattern->getRows();
        }
        
        if (m_pDblConstraintsLinearity->isEmpty() == false && m_pDblConstraintsLinearity->getSize() != m_iNbConstraints)
        {
            Scierror(999,_("%s: Wrong size for input argument #%d (%s): A vector of size %d expected.\n"), getCallerStr(), 
            CONSTR_LIN_TYPE_POS+1, pstrArgName[CONSTR_LIN_TYPE_POS], m_iNbConstraints);
            return false;     
        }
    }
    else // m_iNbConstraints == 0
    {
        if (m_pDblConstraintsJacPattern->isEmpty() == false)
        {
            Scierror(999,_("%s: Wrong size for input argument #%d (%s): Should be an empty matrix.\n"), getCallerStr(), 
            SPARSE_DGOBJ_POS+1, pstrArgName[SPARSE_DGOBJ_POS]);
            return false;             
        }
        if (m_pDblConstraintsLinearity->isEmpty() == false)
        {
            Scierror(999,_("%s: Wrong size for input argument #%d (%s): Should be an empty matrix.\n"), getCallerStr(), 
            CONSTR_LIN_TYPE_POS+1, pstrArgName[CONSTR_LIN_TYPE_POS]);
            return false;             
        }
    }
    
    // Verify Hessian lower triangle sparsity pattern
    if (!in[DHOBJ_POS]->isDouble()) // i.e. is a function or a list
    {
        int iNbMaxTerms = (int) m_iNbVariables*(m_iNbVariables+1.0)/2.0;
        
        if (m_pDblHessianLowerTriPattern->isEmpty())
        {
            m_iNbHessianNonZeroTerms = iNbMaxTerms; // full pattern will be set in OptimizationManager::getHessianLowerTriPattern
        }
        else if (m_pDblHessianLowerTriPattern->getCols() != 2)
        {
            Scierror(999,_("%s: Wrong size for input argument #%d (%s): A matrix with %d columns expected.\n"), getCallerStr(),
            SPARSE_DHOBJ_POS+1, pstrArgName[SPARSE_DHOBJ_POS],2);            
            return false;    
        }
        else if (m_pDblHessianLowerTriPattern->getRows() > iNbMaxTerms)
        {
            Scierror(999,_("%s: Wrong size for input argument #%d (%s): A matrix with at most %d rows expected.\n"), getCallerStr(),
            SPARSE_DHOBJ_POS+1, pstrArgName[SPARSE_DHOBJ_POS], iNbMaxTerms);            
            return false;    
        }
        else
        {
            for (int k=0; k<m_pDblHessianLowerTriPattern->getRows(); k++)
            {
                int iRow = m_pDblHessianLowerTriPattern->get(k,0);
                int iCol = m_pDblHessianLowerTriPattern->get(k,1);
                if (iRow < 1 || iRow > m_iNbVariables || iCol < 1 || iCol > iRow)
                {
                    Scierror(999,_("%s: invalid value in argument #%d (%s) at row %d.\n"), getCallerStr(), 
                    SPARSE_DGOBJ_POS+1, pstrArgName[SPARSE_DHOBJ_POS], k+1);    
                    return false;                            
                }
            }
            m_iNbHessianNonZeroTerms = m_pDblHessianLowerTriPattern->getRows();      
        }
    }
        
    return true;
}

bool OptimizationManager::parseFunction(types::typed_list &in, int iPos, functionKind  what)
{
    types::InternalType *pIn = in[iPos];

    std::string strJacOpt, strHessOpt;
    getIpoptApp()->Options()->GetStringValue("jacobian_approximation", strJacOpt, "");
    getIpoptApp()->Options()->GetStringValue("hessian_approximation", strHessOpt, "");
    
    if (pIn->isDouble() && pIn->getAs<types::Double>()->isEmpty())
    {
        m_pCallFunction[what] = NULL;

        if (what == FOBJ || 
            (what == DGOBJ && strJacOpt == "exact" && in[GOBJ_POS]->isString()) ||
            (what == DHOBJ && strHessOpt == "exact") ||
            (what == DFOBJ && in[GOBJ_POS]->isString()))              
        {
            Scierror(999, _("%s: parameter %d (%s) should be a string, a function or a list.\n"), getCallerStr(), iPos+1, pstrArgName[iPos]);
            return false;
        }
        if (what == DFOBJ) 
        {
            // user function f proto is [f,grad_f] = fun(x)
            m_actualWhat[DFOBJ] = FOBJ;
        }
        else if (what == DGOBJ && in[GOBJ_POS]->isString() == false)
        {
            // user function g proto is [f,jac_g] = fun(x)
            m_actualWhat[DGOBJ] = GOBJ;
        }
    }
    else if (pIn->isCallable())
    {
        m_pCallFunction[what] = pIn->getAs<types::Callable>();
    }
    else if (pIn->isList())
    {
        types::List *pList = pIn->getAs<types::List>();
        types::InternalType *pIFirst =  pList->get(0);
        if (pIFirst->isCallable())
        {
            m_pCallFunction[what] = pIFirst->getAs<types::Callable>();            
        }
        else
        {
            Scierror(999,_("%s: first element of parameter %d (%s) should be a function.\n"), getCallerStr(), iPos+1, pstrArgName[iPos]);
            return false;    
        }
        for (int i=1; i<pList->getSize(); i++)
        {
            m_pParameters[what].push_back(pList->get(i));
        }        
    }
    else if (pIn->isString())
    {
        ConfigVariable::EntryPointStr* func = ConfigVariable::getEntryPoint(pIn->getAs<types::String>()->get(0));
        if (func == NULL)
        {
            Scierror(999,_("%s: unable to find entry point %ls.\n"), getCallerStr(), pIn->getAs<types::String>()->get(0));
            return false;
        }
        else
        {
            m_pEntryPointFunction[what] = func->functionPtr;
        }
    }
    else
    {
        Scierror(999, _("%s: parameter %d (%s) should be a string, a function or a list.\n"), getCallerStr(), iPos+1, pstrArgName[iPos]);
        return false;    
    }

    if (m_pCallFunction[what] != NULL)
    {
        m_pCallFunctionName[what] = wide_string_to_UTF8(m_pCallFunction[what]->getName().c_str());        
    }
        
    return true;
}

bool OptimizationManager::getX0(double *pdblX0)
{
    if (m_pDblX0->isEmpty() == false)
    {
        memcpy(pdblX0, m_pDblX0->get(), sizeof(double)*m_iNbVariables);
        return true;
    }
    return false;
}

bool OptimizationManager::getVariablesLinearity(Ipopt::TNLP::LinearityType* piVarLin)
{
    if (m_pDblVariablesLinearity->isEmpty() == false)
    {
        for (int k=0; k < m_iNbVariables; k++)
        {
            piVarLin[k] = m_pDblVariablesLinearity->get(k) == 0. ? Ipopt::TNLP::LINEAR : Ipopt::TNLP::NON_LINEAR;
        }
        return true;
    }
    return false;
}

bool OptimizationManager::getNonLinearVariables(Ipopt::Index *pos_nonlin_vars)
{
    if (m_piNonLinearVariables != NULL)
    {
        for (int i = 0; i < m_iNbNonLinearVariables; i++)
        {
            pos_nonlin_vars[i] = m_piNonLinearVariables[i];
        }
        return true;        
    }
    return false;
}

bool OptimizationManager::getSolution(double *pdblX, double *pdblLambda, double *pdblZL, double *pdblZU, double *pdblF)
{
    if (m_pdblSolution != NULL)
    {
        memcpy(pdblX, m_pdblSolution, sizeof(double)*m_iNbVariables);
        memcpy(pdblLambda, m_pdblLambda, sizeof(double)*m_iNbConstraints);        
        memcpy(pdblZL, m_pdblZL, sizeof(double)*m_iNbVariables);        
        memcpy(pdblZU, m_pdblZU, sizeof(double)*m_iNbVariables);        
        *pdblF = m_dblFobjSolution;
    }
        
    return true;
}

bool OptimizationManager::getBoundsInfo(double *pdblXLower, double *pdblXUpper, double *pdblConstraintsLhs, double *pdblConstraintsRhs)
{
    int iOne = 1;
    if (m_pDblXLower->isEmpty() == false)
    {
        memcpy(pdblXLower, m_pDblXLower->get(), sizeof(double)*m_iNbVariables);
    }
    else
    {
        std::fill(pdblXLower, pdblXLower+m_iNbVariables, -std::numeric_limits<double>::infinity());
    }
    
    if (m_pDblXUpper->isEmpty() == false)
    {
        memcpy(pdblXUpper, m_pDblXUpper->get(), sizeof(double)*m_iNbVariables);
    }
    else
    {
        std::fill(pdblXUpper, pdblXUpper+m_iNbVariables, std::numeric_limits<double>::infinity());
    }
    
    if (m_pDblConstraintsLhs->isEmpty() == false)
    {
        memcpy(pdblConstraintsLhs, m_pDblConstraintsLhs->get(), sizeof(double)*m_iNbConstraints);
    }
    else
    {
        std::fill(pdblConstraintsLhs, pdblConstraintsLhs+m_iNbConstraints, -std::numeric_limits<double>::infinity());
    }
    
    if (m_pDblConstraintsRhs->isEmpty() == false)
    {
        memcpy(pdblConstraintsRhs, m_pDblConstraintsRhs->get(), sizeof(double)*m_iNbConstraints);
    }
    else
    {
        std::fill(pdblConstraintsRhs, pdblConstraintsRhs+m_iNbConstraints, std::numeric_limits<double>::infinity());
    }
    return true;
}

bool OptimizationManager::getComputed(double *pdbl, functionKind  what)
{
    if (m_pdblOutput[what] != NULL)
    {
        memcpy(pdbl, m_pdblOutput[what], sizeof(double)*m_iSizeOfOutput[what]);
        return true;
    }
    return false;
}

bool OptimizationManager::getConstraintsJacobianPattern(int *piRow, int *piCol)
{
    if (m_pDblConstraintsJacPattern->isEmpty() == false)
    {
        for (int k=0; k < m_iNbConstraintsJacobianNonZeroTerms; k++)
        {
            //pattern is given with Scilab indices (starting from 1)
            piRow[k] = (int) m_pDblConstraintsJacPattern->get(k,0) - 1;
            piCol[k] = (int) m_pDblConstraintsJacPattern->get(k,1) - 1;
        }
        return true;
    }
    else // full jacobian (column-major order)
    {
        int k = 0;
        for (int j = 0; j < m_iNbVariables; j++)
        {
            for (int i = 0; i < m_iNbConstraints; i++)
            {
                piRow[k] = i; 
                piCol[k] = j;
                ++k;
            }                
        }
    }
    return false;
}

bool OptimizationManager::getHessianLowerTriPattern(int *piRow, int *piCol)
{
    if (m_pDblHessianLowerTriPattern->isEmpty() == false)
    {
        for (int k=0; k < m_iNbHessianNonZeroTerms; k++)
        {
            //pattern is given with Scilab indices (starting from 1)
            piRow[k] = (int) m_pDblHessianLowerTriPattern->get(k) - 1;
            piCol[k] = (int) m_pDblHessianLowerTriPattern->get(k+m_iNbHessianNonZeroTerms) - 1;
        }
    }
    else // full hessian 
    {
        int k = 0;
        for (int i = 0; i < m_iNbVariables; i++)
        {
            for (int j = 0; j <= i; j++)
            {
                piRow[k] = i; 
                piCol[k] = j;
                ++k;
            }                
        }        
        m_iNbHessianNonZeroTerms = k;
    }
    return true;
}

bool OptimizationManager::getConstraintsLinearity(Ipopt::TNLP::LinearityType* piConstrLin)
{
    if (m_pDblConstraintsLinearity->isEmpty() == false)
    {
        for (int k=0; k < m_iNbConstraints; k++)
        {
            piConstrLin[k] = m_pDblConstraintsLinearity->get(k) == 1. ? Ipopt::TNLP::LINEAR : Ipopt::TNLP::NON_LINEAR;
        }
        return true;
    }
    return false;
}

bool OptimizationManager::setAllStuff()
{
    m_pdblOutput[FOBJ] = new double[1];
    m_iSizeOfOutput[FOBJ] = 1;
    
    m_pdblOutput[DFOBJ] = new double[m_iNbVariables];
    m_iSizeOfOutput[DFOBJ] = m_iNbVariables;

    // set variables linearity vector + number of non-linear variables
    if (m_pDblVariablesLinearity->isEmpty())
    {
        // all are supposed non-linear
        m_iNbNonLinearVariables = -1; 
    }
    else
    {
        m_piNonLinearVariables = new int[m_iNbVariables];
        m_iNbNonLinearVariables = 0;
        for (int i=0; i<m_iNbVariables; i++)
        {
            if (m_pDblVariablesLinearity->get(i) != 0) // non-linear
            {
                m_piNonLinearVariables[m_iNbNonLinearVariables] = i;
                m_iNbNonLinearVariables++;
            }
        }
    }

    if (m_iNbConstraints > 0)
    {
        m_pdblOutput[GOBJ] = new double[m_iNbConstraints];
        m_iSizeOfOutput[GOBJ] = m_iNbConstraints;
        m_pdblOutput[DGOBJ] = new double[m_iNbConstraintsJacobianNonZeroTerms];
        m_iSizeOfOutput[DGOBJ] = m_iNbConstraintsJacobianNonZeroTerms;
    }
    if (m_pCallFunction[DHOBJ] != NULL)
    {
        m_pdblOutput[DHOBJ] = new double[m_iNbHessianNonZeroTerms];        
        m_iSizeOfOutput[DHOBJ] = m_iNbHessianNonZeroTerms;
    }

    m_pdblSolution  = new double[m_iNbVariables];
    m_pdblLambda  = new double[m_iNbConstraints];
    m_pdblZL = new double[m_iNbVariables];
    m_pdblZU = new double[m_iNbVariables];
    
    return true;
}

OptimizationManager::~OptimizationManager()
{
    for (int i=0; i<5; i++)
    {
        if (m_pdblOutput[i] != NULL)
        {
            delete [] m_pdblOutput[i];
        }
        if (m_pCallFunctionName[i] != NULL)
        {
            delete[] m_pCallFunctionName[i];
        }
    }
    if (m_piNonLinearVariables != NULL)
    {
        delete[] m_piNonLinearVariables;
    }
    if (m_pdblSolution != NULL)
    {
        delete[] m_pdblSolution;
    }
    if (m_pdblLambda != NULL)
    {    
        delete[] m_pdblLambda;
    }
    if (m_pdblZU != NULL)
    {    
        delete[] m_pdblZU;
    }
    if (m_pdblZL != NULL)
    {    
        delete[] m_pdblZL;
    }
}

bool OptimizationManager::computeFunctions(double *x, bool new_x, double obj_factor, double *lambda, bool new_lambda, functionKind  what)
{
    return computeFunctionsAndDerivatives(x, new_x, obj_factor, lambda, new_lambda, what);
}

bool OptimizationManager::computeFunctions(double *x, bool new_x, functionKind  what)
{
    return computeFunctionsAndDerivatives(x, new_x, 0, NULL, false, what);
}

bool OptimizationManager::computeFunctionsAndDerivatives(double *x, bool new_x, double obj_factor, double *lambda, bool new_lambda, functionKind origWhat)
{
    int iOne = 1;
    int iRetCount = 1;
    char errorMsg[256] = "";

    types::typed_list in;
    types::typed_list out;
    types::optional_list opt;
    types::Double* pDblX = NULL;
    types::Bool *pBNewX = NULL;
    types::Double* pDblObjFactor = NULL;
    types::Double* pDblLambda = NULL;
    types::Bool* pBNewLambda = NULL;
    
    functionKind  what = m_actualWhat[origWhat];
    types::Callable *pCall = m_pCallFunction[what];
    
    if (m_bIsComputed[origWhat]) // if already computed (car occur with gradObj or gradConstr)
    {
        return true;
    }

    pDblX = new types::Double(m_iNbVariables, 1);
    memcpy(pDblX->get(), x, sizeof(double)*m_iNbVariables);
    in.push_back(pDblX);

    pBNewX = new types::Bool(new_x);
    in.push_back(pBNewX);

    if (lambda != NULL)
    {
        pDblObjFactor = new types::Double(obj_factor);
        in.push_back(pDblObjFactor);
        pDblLambda = new types::Double(m_iNbConstraints, 1);
        memcpy(pDblLambda->get(), lambda, sizeof(double)*m_iNbConstraints);
        in.push_back(pDblLambda);
        pBNewLambda = new types::Bool(new_lambda);
        in.push_back(pBNewLambda);
    }
    
    // optional user input parameters 
    for (auto pIn : m_pParameters[what])
    {
        in.push_back(pIn);
    }

    for (auto pIn : in)
    {
        pIn->IncreaseRef();
    }

    // when grad_f (resp jac_g) is required but not provided (empty matrix in call) then user function has to return also f (resp g)
    
    m_bIsComputed[what] = true;
    m_bIsComputed[origWhat] = true;
        
    iRetCount = (what == origWhat) ? 1 : 2;
    
    try
    {
        // new std::wstring(L"") is deleted in destructor of ast::CommentExp
        pCall->invoke(in, opt, iRetCount, out, ast::CommentExp(Location(), new std::wstring(L"")));
    }
    catch(ast::InternalError& ie)
    {
        for (auto pIn : m_pParameters[what])
        {
            pIn->DecreaseRef();
        }
        if (ConfigVariable::getLastErrorFunction() == L"")
        {
            // necessary for errors such as "Wrong number of input arguments"
            std::wostringstream ostr;
            ostr << m_pCallFunctionName[what] << L": " <<  ie.GetErrorMessage();
            ie.SetErrorMessage(ostr.str());
        }
        throw(ie);
    }

    for (auto pOut : out) // safety IncreaseRef
    {
        pOut->IncreaseRef();
    }

    for (auto pIn : in)
    {
        pIn->DecreaseRef();
        if (pIn->isDeletable())
        {
            delete pIn;
        }
    }

    if (out.size() != iRetCount)
    {
        sprintf(errorMsg, _("%s: Wrong number of output argument(s): %d expected.\n"), m_pCallFunctionName[what], iRetCount);
        throw ast::InternalError(errorMsg);
    }
    
    functionKind  whatArray[2] = {what, origWhat}; // second element not used if what==origWhat because of iRetCount already set accordingly.
    
    // check type of output and copy to destination if type and dimensions are ok.
    for (int i = 0; i < out.size(); i++)
    {
        if (out[i]->isDouble() == false || out[i]->getAs<types::Double>()->isComplex() || out[i]->getAs<types::Double>()->isEmpty())
        {
            sprintf(errorMsg, _("%s: Wrong type for output argument #%d: Real matrix expected.\n"), m_pCallFunctionName[what], i+1);
            throw ast::InternalError(errorMsg);            
        }
        if (out[i]->getAs<types::Double>()->getSize() != m_iSizeOfOutput[whatArray[i]])
        {
            sprintf(errorMsg, _("%s: Wrong size for output argument #%d: A matrix of size %d expected.\n"), m_pCallFunctionName[what], i+1,m_iSizeOfOutput[whatArray[i]]);   
            throw ast::InternalError(errorMsg);
        }
        memcpy(m_pdblOutput[whatArray[i]], out[i]->getAs<types::Double>()->get(), sizeof(double)*out[i]->getAs<types::Double>()->getSize());                
        // update counter
        m_iNbFunctionCount[whatArray[i]]++;
    }

    for (auto pOut : out)
    {
        pOut->DecreaseRef();
        if (pOut->isDeletable())
        {
            delete pOut;
        }
    }

    return true;
}

bool OptimizationManager::setSolution(double *x, double *lambda, double *z_L, double *z_U, double fobj)
{
    if (m_pdblSolution != NULL)
    {
        memcpy(m_pdblSolution, x, sizeof(double)*m_iNbVariables);
        memcpy(m_pdblLambda, lambda, sizeof(double)*m_iNbConstraints);
        memcpy(m_pdblZL, z_L, sizeof(double)*m_iNbVariables);
        memcpy(m_pdblZU, z_U, sizeof(double)*m_iNbVariables);
        m_dblFobjSolution = fobj;        
    }
    return true;
}

bool OptimizationManager::intermediateCallback(std::vector<std::pair<std::wstring, double>> scalarQuantities, 
     const Ipopt::IpoptData *ip_data, Ipopt::IpoptCalculatedQuantities *ip_cq)
{
    types::typed_list in;
    types::typed_list out;
    types::optional_list opt;
    int iRetCount = 1;
    char errorMsg[256] = "";   
    bool bRet;

    if (m_pCallFunction[INTCB] == NULL)
    {
        return true;
    }
    
    types::Struct *pStructParam = new types::Struct(1,1);

    // get primal and dual variables (if available)
    
    Ipopt::TNLPAdapter* tnlp_adapter = NULL;
    if( ip_cq != NULL )
    {
      Ipopt::OrigIpoptNLP* orignlp;
      orignlp = dynamic_cast<Ipopt::OrigIpoptNLP*>(GetRawPtr(ip_cq->GetIpoptNLP()));
      if( orignlp != NULL)
      {
          tnlp_adapter = dynamic_cast<Ipopt::TNLPAdapter*>(GetRawPtr(orignlp->nlp()));
          
          types::Double *pDblPrimal = new types::Double(m_iNbVariables, 1); 
          types::Double *pDbDualEqs = new types::Double(m_iNbConstraints, 1);
          types::Double *pDbDualLbs = new types::Double(m_iNbVariables, 1);
          types::Double *pDbDualUbs = new types::Double(m_iNbVariables, 1);
          
          tnlp_adapter->ResortX(*ip_data->curr()->x(), pDblPrimal->get());
          tnlp_adapter->ResortG(*ip_data->curr()->y_c(), *ip_data->curr()->y_d(), pDbDualEqs->get());
          tnlp_adapter->ResortBnds(*ip_data->curr()->z_L(), pDbDualLbs->get(),
                                   *ip_data->curr()->z_U(), pDbDualUbs->get());

          pStructParam->addField(L"x");
          pStructParam->get(0)->set(L"x", pDblPrimal);                               
          pStructParam->addField(L"lambda");
          pStructParam->get(0)->set(L"lambda", pDbDualEqs);                               
          pStructParam->addField(L"zL");
          pStructParam->get(0)->set(L"zL", pDbDualLbs);                               
          pStructParam->addField(L"zU");
          pStructParam->get(0)->set(L"zU", pDbDualUbs);                               
      }
    }
    
    // standard quantities
    
    for (int i=0; i<scalarQuantities.size(); i++)
    {
        pStructParam->addField(scalarQuantities[i].first.c_str());        
        types::Double *pDbl = new types::Double(scalarQuantities[i].second);
        pStructParam->get(0)->set(scalarQuantities[i].first.c_str(), pDbl);           
    }
    in.push_back(pStructParam);
    
    // optional user input parameters 
    for (auto pIn : m_pParameters[INTCB])
    {
        in.push_back(pIn);
    }

    for (auto pIn : in)
    {
        pIn->IncreaseRef();
    }    
    
    try
    {
        // new std::wstring(L"") is deleted in destructor of ast::CommentExp
        m_pCallFunction[INTCB]->invoke(in, opt, iRetCount, out, ast::CommentExp(Location(), new std::wstring(L"")));
    }
    catch(ast::InternalError& ie)
    {
      for (auto pIn : in)
      {
          pIn->DecreaseRef();
          if (pIn->isDeletable())
          {
            delete pIn;
          }
      }    
      throw(ie);
    }
    
    out[0]->IncreaseRef(); // safety IncreaseRef
    
    for (auto pIn : in)
    {
        pIn->DecreaseRef();
        if (pIn->isDeletable())
        {
          delete pIn;
        }
    }    

    if (out.size() != 1)
    {
        sprintf(errorMsg, _("%s: Wrong number of output argument(s): %d expected.\n"), m_pCallFunctionName[INTCB], 1);
        throw ast::InternalError(errorMsg);
    }

    if (out[0]->isBool() == false)
    {
        sprintf(errorMsg, _("%s: Wrong type for output argument #%d: boolean expected.\n"), m_pCallFunctionName[INTCB], 1);
        throw ast::InternalError(errorMsg);            
    }
    
    bRet = out[0]->getAs<types::Bool>()->get(0);
        
    out[0]->DecreaseRef();
    if (out[0]->isDeletable())
    {
        delete out[0];
    }
    
    return bRet;        
}
